/*
 * Copyright (c) 1996-1999 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */

/************************************************************************
 * thrtab.h: declaration of ThrTable class
 * 
 * A ThrTable is the thread table for the thread library
************************************************************************/
#ifndef _thread_src_thrtab_h_
#define _thread_src_thrtab_h_


#ifndef THR_TAB_SIZE
#define THR_TAB_SIZE 256
#endif // THR_TAB_SIZE


class ThrTable
{
private:
	struct Entry
	{
		// type of the entry
		item_t	type;

		// value of the entry
		// note that we don't use a union to avoid restrictions
		// on the use of virtual functions on the objects in the
		// unions.  We also avoid a pointer to a base class because
		// we want to avoid dynamic memory allocation.
		Tcb			tcb;
		File		file;
		Socket		socket;
		Signal		sig;
#if defined(i386_unknown_nt4_0)
		WindowsMsgQ	wmsg;
#endif // defined(i386_unknown_nt4_0)

		// in order to take advantage of virtual function mechanisms
		// always points to one of the above value members, corresponding
		// to the type member
		Item*	pval;

		Entry( void ) : type( item_t_unknown ), pval( NULL ) {}
	};


	Entry*		items;
	unsigned	nItems;
	unsigned	next_ndx;

	unsigned	ToIndex( thread_t tid ) const	{ return (tid - 1); }
	thread_t	ToTid( unsigned ndx ) const		{ return (ndx + 1); }

public:
	ThrTable( unsigned int nItems );
	~ThrTable( void );

	thread_t	Alloc( item_t type );
	void		Dealloc( thread_t tid );
	unsigned	Size( void ) const		{ return nItems; }


	item_t	GetType( thread_t tid ) const		{ return items[ToIndex(tid)].type; }
	Item*	GetItem( thread_t tid )				{ return (Item*)(items[ToIndex(tid)].pval); }
	const Item*	GetItem( thread_t tid )	const	{ return (const Item*)(items[ToIndex(tid)].pval); }
	Bindable*	GetBindable( thread_t tid )		{ return (Bindable*)GetItem(tid); }
	const Bindable*	GetBindable( thread_t tid ) const	{ return (const Bindable*)GetItem(tid); }
	IOBindable*	GetIOBindable( thread_t tid )	{ return (IOBindable*)GetItem(tid); }
	Thread	GetThread( thread_t tid )			{ return (Thread)GetItem(tid); }
	File*	GetFile( thread_t tid )				{ return (File*)GetItem(tid); }
	Socket*	GetSocket( thread_t tid )			{ return (Socket*)GetItem(tid); }
#if defined(i386_unknown_nt4_0)
	WindowsMsgQ* GetWindowsMsgQ( thread_t tid )	{ return (WindowsMsgQ*)GetItem(tid); }
#endif // defined(i386_unknown_nt4_0)

	bool	IsValid( thread_t tid )	const		{ unsigned ndx = ToIndex(tid);
												  return (ndx < nItems) && (items[ndx].type != item_t_unknown); }
	bool	IsThread( thread_t tid ) const		{ return (items[ToIndex(tid)].type == item_t_thread); }
	bool	IsFile( thread_t tid ) const		{ return (items[ToIndex(tid)].type == item_t_file); }
	bool	IsSocket( thread_t tid ) const		{ return (items[ToIndex(tid)].type == item_t_socket); }
	bool	IsSignal( thread_t tid ) const		{ return (items[ToIndex(tid)].type == item_t_signal); }
#if defined(i386_unknown_nt4_0)
	bool	IsWindowsMsgQ( thread_t tid ) const	{ return (items[ToIndex(tid)].type == item_t_wmsg); }
#endif // defined(i386_unknown_nt4_0)

	thread_t	FindNext( thread_t tid ) const	{ return ToTid( (ToIndex( tid ) + 1) % nItems ); }
	thread_t	FindNext( thread_t tid, item_t type ) const;
	thread_t	FindNextOwned( thread_t tid, item_t type ) const;

	void	UnbindAll( void );

#if defined(THREADLIB_DEBUG)
	void	WriteTo( ostream& s ) const;
#endif // defined(THREADLIB_DEBUG)

	friend class ThrTableIterator;
};

#if defined(THREADLIB_DEBUG)
ostream& operator<<( ostream& s, const ThrTable& tab );
#endif // defined(THREADLIB_DEBUG)


class ThrTableIterator
{
private:
	const ThrTable& tab;
	unsigned curr_tid;

public:
	ThrTableIterator( const ThrTable& tab );

	bool	Next( thread_t* tidp );
	bool	Next( item_t itype, thread_t* tidp );
	void	Reset( void )						{ curr_tid = 1; }
};


#if defined(THREADLIB_DEBUG)

inline
ostream&
operator<<( ostream& s, const ThrTable& tab )
{
	tab.WriteTo( s );
	return s;
}

#endif // defined(THREADLIB_DEBUG)

#endif // !defined(_thread_src_thrtab_h_)
